/* FILE: avf2ovf.cc                 -*-Mode: c++-*-
 *
 * Program to convert from Any (known) Vector Field file
 * to an ovf vector field file format.
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "oc.h"
#include "nb.h"
#include "vf.h"

/* End includes */

static Vf_OvfDataStyle ods;
static BOOL longformat;
static char *inname;
static Vf_Mesh *mesh;

void Usage(const char* msg=NULL)
{
  if(msg!=NULL) fprintf(stderr,"%s\n",msg);
  fprintf(stderr,"Usage: avf2ovf [-b4|-b8] [-longformat]"
	  " infile >outfile\n");
  if(msg!=NULL) exit(1);
  exit(0);
}

void ProcessArgs(int argc,char **argv)
{
  if(argc<2 || argc>4)
    Usage("Command line error: Wrong number of arguments");

  // Set defaults
  inname=NULL;
  ods=vf_oascii;
  longformat=0;
  // Loop through arguments
  for(int i=1;i<argc;i++) {
    if(argv[i][0]!='-') { // Input file
      if(inname!=NULL) // Error: we already opened an input file
	Usage("Command line error: Multiple input files?");
      inname=argv[i];
    }
    else { // Option
      if(strcmp(argv[i],"-b4")==0)              ods=vf_obin4;
      else if(strcmp(argv[i],"-b8")==0)         ods=vf_obin8;
      else if(strcmp(argv[i],"-longformat")==0) longformat=1;
      else if(strcmp(argv[i],"-h")==0 || strcmp(argv[i],"--help")==0)
	Usage(); // Help request
      else Usage("Command line error: Unknown option");
    }
  }
}

int
Tcl_AppInit(Tcl_Interp *interp)
{
    Oc_AutoBuf ab;
    Oc_SetDefaultTkFlag(0);
    if (Oc_Init(interp) == TCL_ERROR) {
        return TCL_ERROR;
    }
    if (Vf_Init(interp) == TCL_ERROR) {
        return TCL_ERROR;
    }

    // Fill mesh from input file
    Vf_FileInput *vffreader = Vf_FileInput::NewReader(inname);
    if(vffreader==NULL) { // Unknown file type or other error
        fprintf(stderr,"ERROR detected processing file %s\n",inname);
        return Tcl_GlobalEval(interp, ab("exit 1"));
    }
    else { // Reader found for inname
        mesh=vffreader->NewMesh();
        delete vffreader;
        if(mesh==NULL) {
            fprintf(stderr,"ERROR detected while reading file %s\n",inname);
            return Tcl_GlobalEval(interp, ab("exit 2"));
        }
    }

    fprintf(stderr,"Copying Title field: %s\n"
            "You may want to edit this field in the output by hand.\n",
            mesh->GetTitle());

    // Write ovf file to stdout
    Vf_OvfFileOutput ofo;
    ofo.WriteMesh(mesh,stdout,ods,longformat,NULL,NULL);

    // That's all folks!
    fflush(stdout);
    return Tcl_GlobalEval(interp, ab("exit 0"));
}

int Oc_AppMain(int argc,char **argv)
{
    ProcessArgs(argc,argv);
    Oc_Main(argc, argv, Tcl_AppInit);
    return 0;
}

