# FILE: avf2ppm.tcl
#
# Initialization script for avf2ppm
#

package require Oc 1.1
Oc_ForceNoTkDefaultMessage
package require Avf2ppm 1.0

# Name of initialization defaults file
set DEFAULTS_FILE [file join [file dirname [info script]] \
	avf2ppm.def]

# Default option values
set a2pOpts(verbose)  1
set a2pOpts(force)    0
set a2pOpts(config)  {}
set a2pOpts(ipat)    {}
set a2pOpts(opatexp) {(\.[^.]?[^.]?[^.]?$|$)}
set a2pOpts(opatsub) {.ppm}
set a2pOpts(ifiles)  {}
set a2pOpts(ofiles)  {}
set a2pOpts(type)    P6
set a2pOpts(filter)  {}

Oc_AppendUsageInfo "avf2ppm [package provide Avf2ppm]" \
	"Convert vector field files to PPM format" \
[subst {
{-h* {} {Display this help message}}
{-v int {Verbose level; 0=quiet, 1=default}}
{-f {} {Force overwrite on output, even if file exists}}
{-type string\
    {Output PPM type, P3 (ascii) or P6 (binary); default: $a2pOpts(type)}}
{-config string {Plot configuration file}}
{-ipat string {Glob-style match string for input files}}
{-opatexp string {Output regsub exp; default: $a2pOpts(opatexp) }}
{-opatsub string {Output subspec;    default: $a2pOpts(opatsub) }}
{-filter string {Postprocessing output filter}}
{-- {} {End of options}}
{{file ...} {string ...} {Input file(s)}}
}]

proc Usage {{msg {}}} {
    if {![string match {} $msg]} {
	Oc_Log Log $msg error
    }
    Oc_Log Log [Oc_GetUsageInfo] panic
    exit 0
}

# Check for usage request
if {[lsearch -glob $argv "-h*"]>=0} { Usage }

### Default initialization
# Internal configuration array, "just in case" we can't find
# the "defaults" script, DEFAULTS_FILE
array set plot_config {
    colormaps  { Red-Black-Blue Blue-White-Red Teal-White-Red \
            Black-Gray-White White-Green-Black Red-Green-Blue-Red }
    arrow,status       1
    arrow,colormap     Black-Gray-White
    arrow,colorcount   0
    arrow,quantity     z
    arrow,autosample   1
    arrow,subsample    10
    arrow,mag          1
    arrow,antialias    1
    pixel,status       1
    pixel,colormap     Teal-White-Red
    pixel,colorcount   225
    pixel,quantity     x
    pixel,autosample   1
    pixel,subsample    2
    pixel,mag          1
    misc,background    white
    misc,drawboundary  1
    misc,margin        10
    misc,width         640
    misc,height        480
    misc,crop          1
    misc,zoom          0
    misc,rotation      0
}
if {[catch {source $DEFAULTS_FILE} errmsg]} {
    puts stderr \
	    "WARNING: Failure trying to source defaults file $DEFAULTS_FILE"
}

### User specified initialization
proc ArgStrip { index count } {
    global argc argv
    set argv [lreplace $argv $index [expr $index + $count - 1]]
    set argc [llength $argv]
}

# Parse command line, storing results in array a2pOpts
while {[llength $argv]>0} {
    set argsize 1
    set optname {}
    regexp -- {^-(.+)} [lindex $argv 0] dummy optname
    switch -exact -- $optname {
	f { set a2pOpts(force) 1 }
	v {
	    set a2pOpts(verbose) [lindex $argv 1]
	    set argsize 2
	}
	filter -
	opatexp -
	opatsub -
	type {
	    set a2pOpts($optname) [lindex $argv 1]
	    set argsize 2
	}
	config -
	ipat {
	    lappend a2pOpts($optname) [lindex $argv 1]
	    set argsize 2
	}
	- {
	    # End of options; rest are files
	    eval {lappend a2pOpts(ifiles)} [lrange $argv 1 end]
	    set argsize [llength $argv]
	}
	default {
	    # One input file
	    lappend a2pOpts(ifiles) [lindex $argv 0]
	}
    }
    ArgStrip 0 $argsize
}

# Input config files specified on command line, in order.
foreach config_file $a2pOpts(config) {
    source $config_file
}

# Append to file list any files matching -ipat's
foreach pat $a2pOpts(ipat) {
    set temp [lsort -dictionary [glob -nocomplain -- $pat]]
    eval {lappend a2pOpts(ifiles)} $temp
}

# Generate output file list
set a2pOpts(ofiles) {}
foreach ifile $a2pOpts(ifiles) {
    if {[regsub -- $a2pOpts(opatexp) $ifile $a2pOpts(opatsub) ofile]!=1} {
	# Regsub error; Fall back to simple append of substitute string
	set ofile $ifile$a2pOpts(opatsub)
    }
    lappend a2pOpts(ofiles) $ofile
}

if {$a2pOpts(verbose)>=3} {
    # Dump input and output files lists
    puts stderr "##############"
    puts stderr "Input  files: $a2pOpts(ifiles)"
    puts stderr "##############"
    puts stderr "Output files: $a2pOpts(ofiles)"
    puts stderr "##############"
}

if {$a2pOpts(verbose)>=2} {
    puts stderr "ipat: $a2pOpts(ipat)"
    puts stderr "opatexp: $a2pOpts(opatexp)"
    puts stderr "opatsub: $a2pOpts(opatsub)"
}
