# FILE: account.tcl
#
# The server process for an OOMMF account.
#
# It listens on the port name by its first command line argument, and 
# directs communications among all OOMMF threads owned by this account.
#

if {([llength $argv] > 2) || ([llength $argv] == 0)} {
    error "usage: account.tcl <service_port> ?<creator_port>?"
}

global master Omf_export Omf_export_list Omf_export_mtime
set master(version) 0.0

# Set up for auto-loading
set master(relhdir) [string trimright [file dirname [info script]] ./]
set master(abshdir) [file join [pwd] $master(relhdir)]
set master(pdir) [file dirname $master(abshdir)]
set master(libdir) [file dirname $master(pdir)]
set master(root) [file dirname $master(libdir)]
cd $master(root)

if {[lsearch $auto_path $master(libdir)] < 0} {
    lappend auto_path $master(libdir)
}

# Support libraries
package require Oc
package require Net

# Ignore Ctrl-C's, Ctrl-Z's, and child deaths.
Oc_IgnoreInteractiveSignals

# Try to keep going, even if controlling terminal goes down.
Oc_IgnoreTermLoss

# Default timeout = 15 minutes
set master(timeout) 900000
Oc_Option Get Demon timeout master(timeout)

# Log info
set master(log) [em_extension ApplicationLog]

# Init portmap
array set portmap {}

proc FatalError {msg} {
    global master
    $master(log) Append $msg
    if {$master(registered)} {
        Deregister
        return
    }
    exit
}

# Multiple connections share one protocol, and one set of globals
# Define protocol
Net_Protocol New master(protocol) \
        -name [list OOMMF account protocol $master(version)]
$master(protocol) AddMessage start programs {} {
    global Omf_export_list
    if {[CheckExportTimes]} {
	SourceExportFiles  ;# Reread .omfExport.tcl files if necessary
    }
    foreach {id cmd} $Omf_export_list { lappend namelist $id }
    return [list start [concat 0 $namelist]]
}
$master(protocol) AddMessage start launch {program} {
    global Omf_export portmap
    if {[CheckExportTimes]} {
	SourceExportFiles  ;# Reread .omfExport.tcl files if necessary
    }
    if {[info exists Omf_export($program)]} {
        if {[catch {
                eval $Omf_export($program) &
                } threadid]} {
            return [list start [list 1 Launch failed: $threadid]]
        }
# I don't remember why I did this registration on launch.  With the
# new forking behavior of omfsh though, it causes double registration,
# so I've removed it.
#        array set portmap [list $threadid [list $program {}]]
        return [list start [list 0 $threadid]]
    }
    return [list start [list 1 Can't launch $program: no such program]]
}
$master(protocol) AddMessage start threads {} {
    global portmap master
    set ret {}
    foreach thread [array names portmap] {
        if {![string match {} [lindex $portmap($thread) 1]] \
                && ![catch {[lindex $portmap($thread) 2] Class}]} {
            lappend ret [list [lindex $portmap($thread) 0] $thread \
                    [lindex $portmap($thread) 1]]
        }
    }
$master(log) Append "Returning: $ret"
    return [list start [concat 0 $ret]]
}
$master(protocol) AddMessage start lookup {thread} {
    global portmap
    if {[info exists portmap($thread)]} {
        set port [lindex $portmap($thread) 1]
        if {![string match {} $port]} {
            return [list start [list 0 $port]]
        }
    }
    return [list start [list 1 $thread not registered]]
}
$master(protocol) AddMessage start register {thread alias port} {
    global portmap expire_event master
    if {[info exists portmap($thread)]} {
        set existingport [lindex $portmap($thread) 1]
        if {[string match {} $existingport]} {
            set ret [list start [list 0 [array set portmap [list \
                    $thread [lreplace $portmap($thread) 1 2 $port $connection]]]]]
        } else {
            return [list start [list 1 \
                    $thread already registered at port $existingport]]
        }
    } else {
        set ret [list start [list 0 [array set portmap [list $thread \
                [list $alias $port $connection]]]]]
    }
    if {[info exists expire_event] && [array size portmap]} {
        after cancel $expire_event
        unset expire_event
    }
    Oc_EventHandler Generate Oc_Main NewThread
    $master(log) Append "portmap: [array get portmap]"
    return $ret
}
$master(protocol) AddMessage start deregister { thread port } {
    # Ought to call back for security -- and that will require
    # event-driven query-handling.
    global portmap expire_event master
    if {![info exists portmap($thread)]} {
        return [list start [list 1 $thread not registered]]
    }
    if {![string match $port [lindex $portmap($thread) 1]]} {
        return [list start [list 1 permission denied]]
    }
    set ret [list close [list 0 [unset portmap($thread)]]]
    if {![array size portmap] && $master(timeout)} {
        set expire_event [after $master(timeout) [list FatalError \
                "OOMMF account server expired"]]
    }
    $master(log) Append "portmap: [array get portmap]"
    return $ret
}
$master(protocol) AddMessage start portmap {} {
    global portmap master
    if {![info exists portmap]} {
        return [list start [concat 0 {}]]
    }
    return [list start [concat 0 [array get portmap]]]
}

$master(protocol) AddMessage start bye {} {
    return [list close "Bye!"]
}

# A rather ugly hack -- have all connections handle the NewThread event
# by sending out a newthread message.  Would be better to support
# this within the protocol somehow.
Oc_EventHandler New _ Net_Connection Ready [list AddNewThreadHandler %object]
proc AddNewThreadHandler {conn} {
    global master
    $master(log) Append "NewThread handler for connection $conn"
    Oc_EventHandler New _ Oc_Main NewThread [list $conn Tell newthread] \
            -groups [list $conn]
}

# Define and start server 
# add -allow and -deny options to control access
# We have to do a manual registration with host server
set master(registered) 0
Net_Server New master(server) -protocol $master(protocol) \
        -log $master(log) -register 0
set serviceport [lindex $argv 0]
if {[catch {$master(server) Start $serviceport} master(msg)]} {
    FatalError "Can't start OOMMF account server on port $serviceport:
	$master(msg)."
}
set master(port) [$master(server) Port]

# Our account name
set master(account) [Net_Account DefaultAccountName]
# Our home directory
#if {[info exists env(HOME)]} {
#    set master(home) $env(HOME)
#} else {
#    set master(home) [pwd]
#}

# Get export list -- read into global array
set master(exportFiles) {}
#if {[file readable [file join $master(home) .omfExport.tcl]]} {
#    lappend master(exportFiles) [file join $master(home) .omfExport.tcl]
#}
if {[file readable [file join $master(pdir) .omfExport.tcl]]} {
    lappend master(exportFiles) [file join $master(pdir) .omfExport.tcl]
}

proc CheckExportTimes {} {
    # Check .omfExport file modification times against the 
    # modification time when last sourced.  Returns 1 if any
    # of the files have a different mtime, 0 otherwise
    global Omf_export_mtime
    foreach file [array names Omf_export_mtime] {
	if {$Omf_export_mtime($file) != [file mtime $file]} {
	    return 1
	}
    }
    return 0
}

proc SourceExportFiles {} {
    # Reads in .omfExport files
    global master Omf_export Omf_export_list Omf_export_mtime

    catch {unset Omf_export}
    catch {unset Omf_export_list}
    catch {unset  Omf_export_mtime}

    foreach file $master(exportFiles) {
	if {![info exists Omf_export] && ![info exists Omf_export_list]} {
	    $master(log) Append "Sourcing $file ..."
	    if {[catch {uplevel #0 [list source $file]} msg]} {
		$master(log) Append "Error: $msg"
		Oc_Log Log "Error sourcing $file: $msg" warning
	    }
	}
	set Omf_export_mtime($file) [file mtime $file]
    }
    if {![info exists Omf_export] && ![info exists Omf_export_list]} {
	error "Neither Omf_export nor Omf_export_list\
		set in $master(exportFiles)"
    }
    if {![info exists Omf_export]} {
	# Fill Omf_export array from Omf_export_list
	foreach {id action} $Omf_export_list {
	    set Omf_export($id) $action
	}
    }
    if {![info exists Omf_export_list]} {
	# Construct Omf_export_list from Omf_export array
	foreach id [array names Omf_export] {
	    lappend Omf_export_list $id $Omf_export($id)
	}
    }
    $master(log) Append "Exporting: $Omf_export_list"
}

SourceExportFiles

# The host we register with
Net_Host New master(host) -hostname localhost -log $master(log)
Oc_EventHandler New _ $master(host) Delete [list FatalError "host died"]
Oc_EventHandler New _ $master(host) Ready Register -oneshot 1
proc Register {} {
    global master
    $master(log) Append "Sending registration message to host"
    set qid [$master(host) Send register $master(account) $master(port)]
    Oc_EventHandler New _ $master(host) Reply$qid RegisterReply -oneshot 1
    Oc_EventHandler New _ $master(host) Timeout$qid RegisterTimeout
}
proc RegisterTimeout {} {
    FatalError "Timed out waiting to complete registration"
}
proc RegisterReply {} {
    global master
    set reply [$master(host) Get]
    switch -- [lindex $reply 0] {
        0 {
            CallbackCreator
        }
        default {
            after 2000
            CallbackCreator
            FatalError "Registration error: [lrange $reply 1 end]"
        }
    }
}
proc CallbackCreator {} {
    global argv master
    set master(registered) 1
    if {[llength $argv] == 2} {
        # Inform my creator that my server is running.
        set port [lindex $argv 1]
        if {[catch {socket localhost $port} s]} {
            FatalError "Unable to call back $port: $s"
        } else {
            $master(log) Append "Called back $port"
	    catch { puts $s ""} ;# Seems to help close port
	                        ## on other side. -mjd 981009
            close $s
        }
    }
}
proc Deregister {} {
    global master
    $master(log) Append "Sending deregistration message to host"
    if {[catch {
            $master(host) Send deregister $master(account) $master(port)
            } qid]} {
        # Can't send message -- host is probably dead.
        $master(log) Append "Dying without deregistering!"
        exit
    }
    Oc_EventHandler New _ $master(host) Reply$qid DeregisterReply
    Oc_EventHandler New _ $master(host) Timeout$qid DeregisterTimeout
}
proc DeregisterReply {} {
    global master
    set reply [$master(host) Get]
    switch -- [lindex $reply 0] {
        0 {
            $master(log) Append "Deregistered.  Exiting clean."
            exit
        }
        default {
            $master(log) Append \
                "Deregistration error: [lrange $reply 1 end]
	Exiting dirty."
            exit
        }
    }
}
proc DeregisterTimeout {} {
    $master(log) Append "Deregistration timeout.  Exiting dirty."
    exit
}
proc Cleanup {} {
    global portmap expire_event master
    foreach thread [array names portmap] {
        if {[catch {[lindex $portmap($thread) 2] Class}]} {
            # Net_Connection is dead -- remove from portmap
            unset portmap($thread)
        }
    }
    if {![array size portmap] && $master(timeout)} {
        set expire_event [after $master(timeout) \
                [list FatalError "OOMMF account server expired"]]
    }
    # Repeat Cleanup every 20 minutes
    after 1200000 Cleanup
}
Cleanup

vwait master(forever)
