# FILE: application.tcl
#
# TODO: Re-discover why the cache is destroyed each time.
#	Try to devise a scheme where it can be retained.

Oc_Class Oc_Application {

    common platformName

    common path {}

    # Which versions are available for each application name
    array common versions

    # Applications indexed by application name and version
    array common apps

    # Scratch array for partial results of the Resolve / Exec process.
    array common cache

    private common afterId
    ClassConstructor {
        set platformName [[Oc_Config RunPlatform] GetValue platform_name]
        set afterId [after idle [list $class SetSearchPath [list [file \
                dirname [file dirname [file dirname [info script]]]]]]]
    }

    proc SetSearchPath {_path} {
        if {[info exists afterId]} {
            after cancel $afterId
            unset afterId
        }
        if {[catch {eval list $_path} msg]} {
            error "Can't set search path to '$_path'
	Value provided is not a valid Tcl list: $msg"
        }
        # Should we check that each element of the list is a directory?
        set path $_path
        $class Rescan
    }

    proc Rescan {} {
        # Clear away any existing Application Instances
        catch {unset versions}
        catch {unset apps}
        foreach a [$class Instances] {
            $a Delete
        }
        # Source all the appindex.tcl files relative to $path
        # They contain Application Define statements to create
        # Application Instances.
        foreach lib $path {
            foreach file [glob -nocomplain \
                    [file join $lib app * appindex.tcl]] {
                if {[catch {source $file} msg]} {
                    Oc_Log Log "Error sourcing application info file:
$file:
$msg" warning
                }
            }
            foreach file [glob -nocomplain \
                    [file join $lib app * * appindex.tcl]] {
                if {[catch {source $file} msg]} {
                    Oc_Log Log "Error sourcing application info file:
$file:
$msg" warning
                }
            }
        }
        # Add entries for tclsh and wish
        set config [Oc_Config RunPlatform]
        # Tcl 7.5 has problems figuring out the value of
        # TCL_EXEC_PREFIX.  That shouldn't be fatal.
        if {[catch {$config GetValue TCL_EXEC_PREFIX} execPrefix]} {
            return
        }
        set bindir [file join $execPrefix bin]

        # back convert 8.0.3 style patchlevel strings
        regexp {^([0-9]+\.[0-9]+)(.*)$} [info patchlevel] _ verpart rest
        regsub {^\.} $rest p rest
        if {[regexp {[abp]} $rest]} {
            regsub a $rest .0. rest
            regsub b $rest .1. rest
            regsub p $rest .2. rest
        } else {
            set rest .2.0
        }
	# Build candidate list (glob match finds, e.g., cygtclsh80.exe)
        global tcl_platform
	set candidate {}
        if {[string match unix $tcl_platform(platform)]} {
            set file tclsh$verpart
        } elseif {[string match windows $tcl_platform(platform)]} {
            set file tclsh[join [split $verpart .] {}].exe
        }
	lappend candidate [file join $bindir $file]
	lappend candidate [file join $bindir tclsh]
	lappend candidate [glob -nocomplain [file join $bindir *$file]]
	lappend candidate [glob -nocomplain [file join $bindir *tclsh]]
	foreach mytclsh $candidate {
	    if {[file executable $mytclsh]} {
		[$class New _ -name tclsh -version $verpart$rest \
			-machine $platformName -file $mytclsh \
			-directory $bindir] Initialize
		break
	    }
	}
	set candidate {}
        if {[info tclversion] < 8} {
            set verpart [expr {$verpart - 3.4}]
        }
        if {[string match unix $tcl_platform(platform)]} {
            set file wish$verpart
        } elseif {[string match windows $tcl_platform(platform)]} {
            set file wish[join [split $verpart .] {}].exe
        }
	lappend candidate [file join $bindir $file]
	lappend candidate [file join $bindir wish]
	lappend candidate [glob -nocomplain [file join $bindir *$file]]
	lappend candidate [glob -nocomplain [file join $bindir *wish]]
	foreach mywish $candidate {
	    if {[file executable $mytclsh]} {
		[$class New _ -name wish -version $verpart$rest \
			-machine $platformName -file $mywish \
			-directory $bindir] Initialize
		break
	    }
	}
    }

    proc Define {arglist} {
        # Warning: This will double eval.
        if {[string match {} [info script]]} {
            error "'$class Define' must be in an appindex.tcl file"
        }
        set indexFile [Oc_DirectPathname [file join [pwd] [info script]]]
        # Apply syntactic sugar
        set argstring ""
        foreach i $arglist {
            if {![string length $i] || [regexp "\[ \t\r\n]" $i]} {
                append argstring "\{$i\} "
            } else {
                append argstring "$i "
            }
        }
        array set temp [uplevel list $argstring]
        # Filter out any unsupported options
        set args {}
        foreach opt [$class Options * New] {
            catch {lappend args $opt $temp($opt)}
        }
        catch {unset temp}
        array set temp $args
        # Supply additional constructor options
        if {![info exists temp(-directory)]} {
            set temp(-directory) [file dirname $indexFile]
        }
        if {[catch {[eval $class New _ [array get temp]] Initialize} msg]} {
            return -code error $msg
        }
    }

    callback proc CompareVersions {v1 v2} {
        while {1} {
            regexp {^([0-9]+)?(\.(.*))?} $v1 match1 num1 _ v1
            if {[set done1 [string match {} $match1]]} {
                set num1 0 
            }
            regexp {^([0-9]+)?(\.(.*))?} $v2 match2 num2 _ v2
            if {[set done2 [string match {} $match2]]} {
                set num2 0
            }
            if {$num1 > $num2} {
                return 1
            } elseif {$num1 < $num2} {
                return -1
            }
            if {$done1 && $done2} {
                break
            }
        }
        return 0
    }

    proc Exec {keys args} {
        if {[catch {eval $class SelectAppByKeys $keys} app]} {
            return -code error $app
        }
        set code [catch {eval $app Exec $args} msg]
        catch {unset cache}
        if {$code} {
            set emsg "Error executing [$app Cget -name] version\
                    [$app Cget -version]:\n\t$msg"
            error $emsg $emsg
        }
        return $msg
    }

    proc CommandLine {keys args} {
        if {[catch {eval $class SelectAppByKeys $keys} app]} {
            return -code error $app
        }
        set code [catch {$app CommandLine} msg]
        catch {unset cache}
        if {$code} {
            set emsg "Error determining [$app Cget -name] version\
                    [$app Cget -version] command line:\n\t$msg"
            error $emsg $emsg
        }
        return [concat $msg $args]
    }

    private proc SelectAppByKeys {args} {
        if {[catch {eval $class Resolve $args} app]} {
            set avail {}
            foreach a [$class Instances] {
                if {[$a WillRun]} {
                    lappend avail "[$a Cget -name] [$a Cget -version]"
                }
            }
            set navail {}
            set last ""
            foreach now [lsort $avail] {
                if {[string compare $last $now]} {
                    lappend navail $now
                    set last $now
                }
            }
            if {[llength $navail]} {
                set avmsg "Available applications for platform\
                        '$platformName':\n\t[join $navail \n\t]"
            } else {
                set avmsg "Have you run pimake for platform '$platformName'?"
            }
            catch {unset cache}
            return -code error "Can't find application satisfying\
                    '$args'\nfor platform '$platformName':\n\t$app\n\n$avmsg"
        }
        return $app
    }

    private proc Resolve {args} {
        # This calls SetSearchPath in the event that the event
        # loop still hasn't had a chance to evaluate the SetSearchPath
        # call scheduled by the Class Constructor
        if {[info exists afterId]} {
            uplevel #0 [lindex [after info $afterId] 0]
        }
        if {[llength $args] == 0} {
            return ""
        }
        set name [string tolower [lindex $args 0]]
        if {[string match {} $name]} {
            return -code error "No application name specified."
        }
        set args [lrange $args 1 end]
        if {[set exact [string match -exact [lindex $args 0]]]} {
            set args [lrange $args 1 end]
        }
        set version [lindex $args 0]
        if {$exact} {
            if {[string match {} $version]} {
                return -code error "'-exact' option requires a version number"
            }
            if {![info exists apps($name,$version)]} {
                return -code error "No matching application in the search path."
            }
            set attempts ""
            foreach app $apps($name,$version) {
                if {[$app WillRun]} {
                    return $app
                }
                append attempts "$cache($app)\n"
            }
            return -code error "The following candidates were\
                    rejected:\n$attempts"
        }
        if {![info exists versions($name)]} {
            return -code error "No matching application in the search path."
        }
        set attempts ""
        foreach v [lsort -command [list $class CompareVersions] -decreasing \
                $versions($name)] {
            if {[$class Satisfies $v $version] \
                    && [info exists apps($name,$v)]} {
                foreach app $apps($name,$v) {
                    if {[$app WillRun]} {
                        return $app
                    }
                    append attempts "$cache($app)\n"
                }
            }
        }
        return -code error "The following candidates were rejected:\n$attempts"
    }

    private proc Satisfies {avail req} {
        if {[string match {} $req]} {
            return 1
        }
        regexp {^([0-9]+)?(\.(.*))?} $avail _ ma
        regexp {^([0-9]+)?(\.(.*))?} $req _ mr
        expr {($ma == $mr) && ([$class CompareVersions $avail $req] >= 0)}
    }

    const public variable machine
    const public variable name 
    const public variable version
    const public variable file

    const public variable aliases = {}
    const public variable directory

    Constructor {args} {
        eval $this Configure $args
    }

    method Initialize {} {
        foreach _ [concat [list $name] $aliases] {
            set n [string tolower $_]
            lappend apps($n,$version) $this
            if {![info exists versions($n)] \
                    || [lsearch -exact $versions($n) $version] < 0} {
                lappend versions($n) $version
            }
        }
        return {}
    }

    # execute this Application
    private method Exec {args} {
        eval $cache($this) $args
    }

    private method CommandLine {} {
        set firstword [lindex $cache($this) 0]
        if {[string match exec $firstword]} {
            return [lrange $cache($this) 1 end]
        }
        return [concat [$firstword CommandLine] [lrange $cache($this) 2 end]]
    }
    
    private method WillRun {} {
        set execfile [file join $directory $file]
        # Try to resolve the machine into a supporting shell application
        if {![catch {eval $class Resolve $machine} app]} {
            # Found a shell app which WillRun
            if {[file readable $execfile]} {
                set cache($this) [list $app Exec $execfile]
                return 1
            } else {
                set cache($this) "$execfile: permission denied"
                return 0
            }
        }
        # No shell app -- try as a binary
        if {[string match $platformName $machine]} {
            if {[file executable $execfile]} {
                set cache($this) [list exec $execfile]
                return 1
            } else {
                set cache($this) "$execfile: permission denied"
                return 0
            }
        } else {
            set cache($this) "$execfile: needs shell or platform '$machine'"
            return 0
        }
    }
       
}

